# (c) Alexander Belchenko, 2006, 2007

"""Property Panel Factory.

Universal class for creating various custom properties panels.

Each property panel could have several label with editing entry.
E.g.::

    +-----------------------------------------+
    |      id:[ logo                        ] |
    |    name:[ Logo                        ] |
    +-----------------------------------------+

Each element packed to grid with stretch right side and fixed-size
left side.
"""

__docformat__ = "restructuredtext"


# IMPORTS
import logging

from Tkinter import *
from tkColorChooser import askcolor
from tkFileDialog import askopenfilename

from group_dialog import GroupDialog


# HELPER CLASSES
class _PropertiesContainer:
    """Helper class"""

    def add(self, property_name, obj):
        self.__dict__["_obj_" + property_name] = obj

    def __getattr__(self, name):
        return self.__dict__["_obj_" + name].get()

    def __setattr__(self, name, value):
        self.__dict__["_obj_" + name].set(value)


# PUBLIC CLASSES
class PropertyPanel(Frame):
    """Property Panel Factory"""

    def __init__(self, properties, parent=None, options=None):
        """Create new panel with required set of properties.

        :param  properties:     sequence of properties.
                                Each property defined as 3-tuple:
                                (text label of property, class that used
                                to create right side editable/readonly
                                element, default value). Last component
                                is optional, i.e. you can omit default
                                value. Then None will be used as default.

        :type   properties:     sequence of 3-tuple or 2-tuple.

        :param  parent:         master widget for created Frame.

        :param  options:        dict with options for creating new Frame.
        :type   options:        dict or None.

        For each property element will be created corresponding
        object.property with name as in label of property.
        object.property linked to setters/getters
        of created class with names .get/.set
        """
        if options is None:
            Frame.__init__(self, parent, relief=GROOVE, borderwidth=2)
        else:
            Frame.__init__(self, parent, **options)
        self.parent = parent
        self._first_item = None  # first item that should receive input focus

        assert len(properties) > 0, "Sequence of properties shouldn't be empty"

        __ = _
        labels_max_len = 0
        for prop in properties:
            label = prop[0]
            if not label.strip():
                continue
            w = len(__(label))
            labels_max_len = max(labels_max_len, w)
#            logging.debug('PropertyPanel: label=%r, w=%d, max_len=%d', prop[0], w, labels_max_len)
        labels_max_len += 1

        self.property = _PropertiesContainer()
        self.property_object = {}

        for ix, prop in enumerate(properties):
            # unpack property tuple
            length = len(prop)
            if length == 2:
                label, klass = prop
                default = None
            elif length == 3:
                label, klass, default = prop
            elif length == 4:
                label, klass, items, default = prop
            else:
                raise 'Incorrect length of property %r' % prop

            # make label
            if label.strip():
                label_text = __(label) + ':'
            else:
                label_text = label
            Label(self, text=label_text,
                  anchor=E, width=labels_max_len).grid(column=0, row=ix)
            # make edit entry object
            if length < 4:
                obj = klass(self, default)
            else:
                obj = klass(self, items, default)
            obj.name = label
            obj.grid(column=1, row=ix, padx=4, pady=2, sticky=E+W+N+S)
            if obj.weight:
                self.rowconfigure(ix, weight=obj.weight)
            self.property.add(label, obj)
            self.property_object[label] = obj
            if self._first_item is None:
                self._first_item = obj

        self.columnconfigure(1, weight=1)

    def get_first_item(self):
        """Return first item on property panel that should have focus"""
        return self._first_item

#/class PropertyPanel


class IProperty:
    """Interface for property-derived classes"""

    weight = None

    def __init__(self, parent, default=None):
        """Interface for constructor method.
        :param   parent:    master widget for this class
        :param   default:   default value for property
        """
        raise NotImplementedError

    def get(self):
        """Getter of current property value"""
        raise NotImplementedError

    def set(self, value):
        """Setter of new value to property"""
        raise NotImplementedError

#/class IProperty


class PropertyEntry(IProperty, Entry):
    """Tkinter.Entry-based property editing"""

    def __init__(self, parent, default=None):
        Entry.__init__(self, parent, width=30)
        self._var = StringVar()
        self.configure(textvariable=self._var)

        self.set(default)

    def set(self, value):
        if value is None:
            self._var.set("")
            self["state"] = DISABLED
        else:
            self._var.set(value)
            self["state"] = NORMAL

    def get(self):
        return self._var.get()
#/class PropertyEntry


class PropertyEntryInt(PropertyEntry):
    """Property Entry for edit integer values"""

    def get(self):
        try:
            return int(self._var.get())
        except ValueError:
            return 0
#/class PropertyEntryInt


class PropertyEntryWithSelect(IProperty, Frame):
    """Base class for creating complex widgets that behaves
    as entry with button at right.
    Derived classes should define .onSelect method
    """
    def __init__(self, parent, default=None):
        Frame.__init__(self, parent, borderwidth=0)

        self._btn = Button(self, text="...", command=self.onSelect)
        self._btn.pack(side=RIGHT)

        self._var = StringVar()
        self._entry = Entry(self, textvariable=self._var, width=30)
        self._entry.pack(side=LEFT, fill=BOTH, expand=YES)

        self.set(default)

    def onSelect(self):
        """Callback method invoked when user press button"""
        raise NotImplementedError

    def get(self):
        return self._var.get()

    def set(self, value):
        if value is None:
            self._var.set("")
            self._entry["state"] = DISABLED
            self._btn["state"] = DISABLED
        else:
            self._var.set(value)
            self._entry["state"] = NORMAL
            self._btn["state"] = NORMAL
#/class PropertyEntryWithSelect


class PropertySpacer(IProperty, Frame):
    """Simple spacer as empty label"""

    def __init__(self, parent, default=None):
        Frame.__init__(self, parent, borderwidth=0)
        Label(self, text=default or ' ').pack(fill=BOTH, expand=YES)
#/class PropertySpacer


class PropertyLabelWithSelect(IProperty, Frame):
    """Base class for creating complex widgets that behaves
    as read-only label with button at right. Editing of this entry
    should be performed only via additional dialog invoked by
    select button.
    Derived classes should define .onSelect method
    """
    def __init__(self, parent, default=None):
        Frame.__init__(self, parent, borderwidth=0)

        self._btn = Button(self, text="...", command=self.onSelect)
        self._btn.pack(side=RIGHT)

        self._var = StringVar()
        self._label = Label(self, textvariable=self._var, width=30,
                            relief=SUNKEN, bd=2, anchor=W)
        self._label.pack(side=LEFT, fill=BOTH, expand=YES)

        self.set(default)

    def onSelect(self):
        """Callback method invoked when user press button"""
        raise NotImplementedError

    def get(self):
        return self._var.get()

    def set(self, value):
        if value is None:
            self._var.set("")
            self._btn["state"] = DISABLED
        else:
            self._var.set(value)
            self._btn["state"] = NORMAL
#/class PropertyLabelWithSelect


class PropertyGroupWithChangeDialog(IProperty, Frame):
    """This class connects GroupDialog from group_dialog.py
    to Label and [Change] button.
    Similar to PropertyLabelWithSelect
    """
    def __init__(self, parent, items_list, default):
        Frame.__init__(self, parent, borderwidth=0)

        self.items_list = items_list

        text = _("Change")
        w = 2 + len(text)
        self._btn = Button(self, text=text, width=w, command=self.onSelect)
        self._btn.pack(side=RIGHT)

        self._var = StringVar()
        self._label = Label(self, textvariable=self._var, width=20,
                            relief=SUNKEN, bd=2, anchor=W)
        self._label.pack(side=LEFT, fill=BOTH, expand=YES)

        self.set(default)

    def onSelect(self):
        """Callback method invoked when user press button"""
        d = GroupDialog(self, self.name, self.items_list, self.get())
        if d.result:
            self.set(d.result)

    def get(self):
        return self.value or ''

    def set(self, value):
        if value is None:
            self._var.set("")
            self._btn["state"] = DISABLED
        else:
            __ = _
            self._var.set(__(value))
            self._btn["state"] = NORMAL
        self.value = value
        
#/class PropertyLabelWithChangeDialog


class PropertyColor(PropertyEntryWithSelect):
    """Complex widget to choose color with standard dialog"""

    def onSelect(self):
        initial_color = self.get()
        if initial_color == "None":
            initial_color = ""

        color_hex = askcolor(initial_color)[1]
        if color_hex == None:
            return

        self.set(color_hex.upper())
#/class PropertyColor


class PropertyEntryFile(PropertyEntryWithSelect):
    """Complex widget to choose filename with standard dialog.
    Also allows to edit filename by hand.
    """
    def __init__(self, parent, default=None):
        PropertyEntryWithSelect.__init__(self, parent, default)
        self._entry.configure(width=40)

    def onSelect(self):
        # ask source file
        filename = askopenfilename()
        if filename in ("", None):
            return
        self.set(filename)
#/class PropertyEntryFile


class PropertyLabelFile(PropertyLabelWithSelect):
    """Complex widget to choose filename only with standard dialog"""

    def __init__(self, parent, default=None):
        PropertyLabelWithSelect.__init__(self, parent, default)
        self._label.configure(width=40)

    def onSelect(self):
        # ask source file
        filename = askopenfilename()
        if filename in ("", None):
            return
        self.set(filename)
#/class PropertyLabelFile


class PropertyRadiobuttonList(IProperty, Frame):
    """Small list of groupped Radiobuttons to select only one variant.
    Values behind each radio button is integers indexes started from 0.
    """
    def __init__(self, parent, items_list, default=0):
        Frame.__init__(self, parent, relief=GROOVE, borderwidth=2)

        self._var = IntVar()

        for ix, item in enumerate(items_list):
            Radiobutton(self, text=item, variable=self._var,
                        value=ix).pack(anchor=W, side=TOP)

        self.get = self._var.get
        self.set = self._var.set

        self.set(default)
#/class PropertyRadiobuttonList


class PropertyRadiobuttonStringList(IProperty, Frame):
    """Small list of groupped Radiobuttons to select only one variant.
    Values behind each radio button is strings equal to shown text.
    """
    def __init__(self, parent, items_list, default=None):
        Frame.__init__(self, parent, relief=GROOVE, borderwidth=2)

        self._var = StringVar()
        for item in items_list:
            Radiobutton(self, text=item, variable=self._var,
                        value=item).pack(anchor=W, side=TOP)

        self.get = self._var.get
        self.set = self._var.set

        if default is None:
            default = items_list[0]
        self.set(default)
#/class PropertyRadiobuttonStringList


class PropertyText(IProperty, Frame):
    """Text editing with scrool bars."""

    weight = 1

    def __init__(self, parent, default):
        Frame.__init__(self, parent)

        sbarv = Scrollbar(self, orient=VERTICAL)
        sbarv.pack(side=RIGHT, fill=Y)
        sbarh = Scrollbar(self, orient=HORIZONTAL)
        sbarh.pack(side=BOTTOM, fill=X)

        text = Text(self, relief=SUNKEN,
                    width=40, height=13,
                    padx=2, pady=2,
                    wrap=NONE)
        text.pack(fill=BOTH, expand=YES)
        self.text = text

        sbarv.config(command=text.yview)
        sbarh.config(command=text.xview)
        text.config(yscrollcommand=sbarv.set, xscrollcommand=sbarh.set)

        self.set(default)

    def get(self):
        return self.text.get('1.0', END)

    def set(self, value):
        self.text.delete('1.0', END)
        if value is None:
            self.text["state"] = DISABLED
        else:
            self.text["state"] = NORMAL
            self.text.insert('1.0', value)
            self.text.mark_set(INSERT, '1.0')
            #self.text.focus()
#/class PropertyText

# MAIN
if __name__ == "__main__":
    root = Tk()

    properties = [('spam', PropertyEntry, 'haha'),
                 ]
    pp = PropertyPanel(properties, root, options={'relief': FLAT})
    pp.pack(side=TOP, fill=X)

    properties = [('foo', PropertyEntry, 'bar'),
                  ('disabled', PropertyEntry),
                  ('entry_int', PropertyEntryInt, -10),
                  ('entry_w_select', PropertyEntryWithSelect),
                  ('entry_w_select2', PropertyEntryWithSelect, ''),
                  ('select_color', PropertyColor, 'Green'),
                  ('ro_entry_select', PropertyLabelWithSelect, 'Test'),
                  ('', PropertySpacer),
                  ('radio', PropertyRadiobuttonList, ['red', 'green', 'blue'], 1),
                  ('radioS', PropertyRadiobuttonStringList, ['long', 'short', 'line'], 'long'),
                  ('group', PropertyGroupWithChangeDialog, ['default', 'long', 'short', 'line'], 'line'),
                  ('entry_file', PropertyEntryFile, 'file.ext'),
                  ('label_file', PropertyLabelFile, 'file.ext'),
                  ('text', PropertyText, ''),
                 ]
    pp = PropertyPanel(properties, root)
    pp.pack(expand=YES, fill=BOTH)

    def get_pp():
        for item in properties:
            label = item[0]
            if label.strip():
                print label, '=>', repr(getattr(pp.property, label))

    Button(root, text="get", command=get_pp).pack()

    root.mainloop()
