"""Model of options data

General options
===============
Corresponds to [DEFAULT] section of bazaar.conf

editor                  name-of-program

email                   Your Name <your@email.address>

check_signatures        require|ignore|check-available(default)

create_signatures       always|never|when-required(default)

gpg_signing_command     name-of-program

log_format              name-of-format

explanation of options
----------------------
editor - this option sets the pop up editor to use during commits.
email - this option sets the user id bzr will use when committing.
check_signatures - this option controls whether bzr will require good gpg
                   signatures, ignore them, or check them if they are 
                   present.
create_signatures - this option controls whether bzr will always create 
                    gpg signatures, never create them, or create them if the
                    branch is configured to require them.
log_format - this option sets the default log format.  Possible values are
             long, short, line, or a plugin can register new formats.

Edit general options
--------------------
email split to 2 text entry: name and e-mail.

editor and gpg_signing_command: entry with ability to select filename
via standard file open dialog.

check_signatures and create_signatures: combobox or group of radiobuttons.
(Does this options should be accessible only when gpg_signing_command
is set?).

log_format: combobox or group of radiobuttons


Aliases options
===============
Each alias has the name and a string with new command line string.
To edit existing alias:
- you can change name of alias
- you can edit command string of alias

To add new alias:
- press button New and get empty alias entry
- edit this alias entry as described above for existing alias

To delete existing alias:
- each alias should have associated button 'Delete'
- user press this button and alias is gone


Plugins manager
===============
For plugins we want to disable/enable existing plugins by moving
them between default plugins folder (i.e. $BZR_PLUGIN_PATH or
./bazaar/plugins) and some additional folder (to keep disabled plugins
in touch).
Also we need the service to obtain and install new plugins from
http://bazaar-vcs.org/PluginRegistry. Open this page with
suffix '?action=raw' we can obtain wiki text source of page
that will easy to parse. Having the list of plugins in registry
we could provide option to user to download or make checkout
(lightweight checkout by default) of some plugins.

Also we can provide option to delete plugin completely.


Branch options
==============
Change options associated with current branch. Available only
when bzrconf launched from directory containing bzr branch.
This page allows to edit options similar to general section,
but also: nick of branch, push/pull/submit location?
"""

import logging
import os

from bzrlib import (config,
                    log,
                   )


# OPTIONS types
ENTRY = 'entry'         # simple editable entry
ENTRY2 = 'entry2'       # pair of editable entries
FILENAME = 'filename'   # entry with additional filename selector
EDITOR_DISABLED = 'editor_disabled'     # when user has $BZR_EDITOR set
GROUP = 'group'         # radiogroup or combobox with predefined variants
SPACER = 'spacer'


# for gettext
# some strings should use deffered translation method
# to simplify this we provide fake method
_ = lambda x: x


##
# get global config instance. let's keep the same instance as singleton.
_global_config = None

def get_global_config():
    global _global_config
    if _global_config is None:
        _global_config = config.GlobalConfig()
    return _global_config


class BaseModel(object):
    """Base model class"""

    name = None

    def __init__(self):
        pass

    def save(self):
        """Save changes in options back to config file"""
        pass

    def reset(self):
        """Cancel non-saved changes in options and reload old values"""
        pass
#/class BaseModel


class GeneralModel(BaseModel):
    """General options model"""

    name = _('General')

    def __init__(self):
        self._config = get_global_config()
        self.fields = [UsernameField(),
                       SpacerField(),
                       EditorField(),
                       SpacerField(),
                       GpgProgramField(),
                       CheckSignaturesField(),
                       CreateSignaturesField(),
                       SpacerField(),
                       LogFormatField(),
                      ]

    def save(self):
        for i in self.fields:
            i.save()

    def reset(self):
        for i in self.fields:
            i.reset()

    def __iter__(self):
        return iter(self.fields)

    def __repr__(self):
        items = []
        for i in self:
            if i.name:
                items.append('%s=%r' % (i.name, i.value))
        return 'GeneralModel(%s)' % ','.join(items)
#/class GeneralModel


class AliasesModel(BaseModel):
    """Aliases for Bazaar"""

    name = _('Aliases')

    def __init__(self):
        self._config = get_global_config()
        parser = self._config._get_parser()
        self.fields = []
        if 'ALIASES' in parser:
            aliases = parser['ALIASES']
            for a in aliases:
                self.fields.append(AliasField(a, aliases[a]))
#        logging.debug('Aliases: %r', self.fields)
        self._old = self.fields[:]

    def __getitem__(self, index):
        """Return (alias, command)"""
        return self.fields[index]()

    def __setitem__(self, index, value):
        """Set new (alias, command) at index"""
        alias, command = value
        if index == len(self.fields):
            self.fields.append(None)
        if index < len(self.fields):
            self.fields[index] = AliasField(alias, command)
        else:
            raise IndexError, 'Index %d too big' % index

    def __delitem__(self, index):
        self.fields.pop(index)

    def __len__(self):
        return len(self.fields)

    def __iter__(self):
        return iter(self.fields)

    def insert(self, index, value):
        """Insert new (alias, command) before index"""
        if index < len(self.fields):
            self.fields.insert(index, None)
        self[index] = value

    def save(self):
#        logging.debug('Aliases to save: %r', self.fields)
        if self.fields != self._old:
            parser = self._config._get_parser()
            if 'ALIASES' in parser:
                aliases = parser['ALIASES']
                for a in aliases.keys():
                    del aliases[a]
            else:
                parser['ALIASES'] = {}
            aliases = parser['ALIASES']
            for a in self:
                aliases[a.alias] = a.command
            f = open(self._config._get_filename(), 'wb')
            parser.write(f)
            f.write('\n')
            f.close()
            self._old = self.fields[:]

    def reset(self):
        self.fields = self._old[:]
#/class AliasesModel


# Fields data objects
class GeneralField(object):
    """General field base object"""

    name = None
    kind = None
    value = None
    variants = None

    def __init__(self):
        self._config = get_global_config()
        self.init()

    def _save_or_delete_option(self, option, value):
        if value:
            self._config.set_user_option(option, value)
        else:
            self._delete_option(option)
                
    def _delete_option(self, option):
        conf = self._config
        conf_file = conf._get_filename()
        if not os.path.isfile(conf_file):
            return
        parser = conf._get_parser()
        if 'DEFAULT' not in parser:
            return
        if parser['DEFAULT'].get(option) is not None:
            del parser['DEFAULT'][option]
            f = open(conf_file, 'wb')
            parser.write(f)
            f.write('\n')
            f.close()

    # override following methods
    def init(self):
        pass

    def save(self):
        pass

    def reset(self):
        pass
#/class GeneralField


class SpacerField(GeneralField):
    name = ''
    kind = SPACER
    value = None
#/class SpacerField


class UsernameField(GeneralField):
    name = (_('Name'), _('E-mail'))
    kind = ENTRY2

    def init(self):
        username = self._config.username()
        try:
            e_mail = self._config.user_email()
        except config.errors.NoEmailInUsername:
            name = username
            e_mail = ''
        else:
            ix = username.index(e_mail)
            if ix > 0:
                if username[ix-1:ix] == '<':
                    ix = ix - 1
                name = username[:ix].strip()
            else:
                ix = e_mail.index('@')
                name = e_mail[:ix-1]
        self.value = (name, e_mail)
        self._old = self.value[:]

    def reset(self):
        self.value = self._old[:]

    def save(self):
        if self.value != self._old:
            name, e_mail = [i.strip() for i in self.value]
            username = '%s <%s>' % (name, e_mail)
            self._save_or_delete_option('email', username)
            self._old = (name, e_mail)
#/class UsernameField


class EditorField(GeneralField):
    name = _('Editor')
    kind = FILENAME

    def init(self):
        bzr_editor_env = os.environ.get('BZR_EDITOR')
        if bzr_editor_env:
            self.value = bzr_editor_env
            self.kind = EDITOR_DISABLED
        else:
            editor_env = ''
            for varname in ('VISUAL', 'EDITOR'):
                if varname in os.environ:
                    editor_env = os.environ[varname]
            self.value = self._config.get_editor() or editor_env
        self._old = self.value

    def reset(self):
        self.value = self._old

    def save(self):
        if self.value != self._old:
            self._save_or_delete_option('editor', self.value)
            self._old = self.value
#/class EditorField


class GpgProgramField(GeneralField):
    name = _('GPG Program')
    kind = FILENAME

    def init(self):
        self.value = self._config.gpg_signing_command() or ''
        self._old = self.value

    def reset(self):
        self.value = self._old

    def save(self):
        if self.value != self._old:
            self._save_or_delete_option('gpg_signing_command', self.value)
            self._old = self.value
#/class GpgProgramField


class CheckSignaturesField(GeneralField):
    name = _('Check Signatures')
    kind = GROUP
    variants = (_('default'),_('require'),_('ignore'),_('check-available'))

    def init(self):
        self.value = (self._config._get_user_option('check_signatures')
                      or 'default')
        self._old = self.value

    def reset(self):
        self.value = self._old

    def save(self):
        if self.value != self._old:
            if self.value == 'default':
                value = None
            else:
                value = self.value
            self._save_or_delete_option('check_signatures', value)
            self._old = self.value
#/class CheckSignaturesField


class CreateSignaturesField(GeneralField):
    name = _('Create Signatures')
    kind = GROUP
    variants = (_('default'),_('always'),_('never'),_('when-required'))

    def init(self):
        self.value = (self._config._get_user_option('create_signatures')
                      or 'default')
        self._old = self.value

    def reset(self):
        self.value = self._old

    def save(self):
        if self.value != self._old:
            if self.value == 'default':
                value = None
            else:
                value = self.value
            self._save_or_delete_option('create_signatures', value)
            self._old = self.value
#/class CreateSignaturesField


class LogFormatField(GeneralField):
    name = _('Log format')
    kind = GROUP
    variants = ['default']

    def init(self):
        for name in ('FORMATTERS', 'log_formatter_registry'):
            fmt = getattr(log, name, None)
            if fmt:
                keys = sorted(fmt.keys())
                break
        else:
            keys = []
        self.variants.extend(keys)
        lfmt = self._config._log_format() or 'default'
        if lfmt not in self.variants:
            lfmt = 'default'
        self.value = lfmt
        self._old = self.value

    def reset(self):
        self.value = self._old

    def save(self):
        if self.value != self._old:
            if value == 'default':
                value = None
            else:
                value = self.value
            self._save_or_delete_option('log_format', value)
            self._old = self.value
#/class LogFormatField


class AliasField(object):
    """Alias field object"""

    def __init__(self, alias, command):
        self.alias = alias
        self.command = command

    def __call__(self):
        return (self.alias, self.command)

    def __repr__(self):
        return 'AliasField(%r,%r)' % (self.alias, self.command)
#/class AliasField
